$(function(){

  //utility
  function loadJSX() {
    var csInterface = new CSInterface()
    , extensionRoot = csInterface.getSystemPath(SystemPath.EXTENSION) + "/jsx/";
    csInterface.evalScript('$._ext.evalFiles("' + extensionRoot + '")');
  }

  function encode(str) {
    return encodeURIComponent(str).replace(/[!'()]/g, escape).replace(/\*/g, "%2A");
  }

  function encodeData(data) {
    return encode(JSON.stringify(data));
  }

  function htmlspecialchars(str) {
    return str.replace(/&/g,"&amp;").replace(/"/g,"&quot;").replace(/'/g,"&#039;").replace(/</g,"&lt;").replace(/>/g,"&gt;");
  }

  function zfill(number, width) {
    function f(number, width) {
      return ((new Array(width + 1)).join('0') + number).slice(-width);
    }
    return number >= 0 ? f(number, width) : '-' + f(-number, width - 1);
  }

  function clamp(v, mn, mx) {
    return v < mn ? mn : v > mx ? mx : v;
  }

  var StorageManager = (function () {
    var localStorage = window.localStorage
    , white_lists = ['__code__', '__folder__', '__subfolder__', '__file_name__'];

    function convert(str) {
      return '__' + str + '__';
    }

    var StorageManager = {};

    StorageManager.clear = function () {
      for (var key in localStorage) {
        if (white_lists.indexOf(key) == -1) {
          delete localStorage[key];
        }
      }
    };

    StorageManager.is = function (key) {
      key = convert(key);
      return localStorage[key] !== undefined;
    };

    StorageManager.set = function (key, value) {
      key = convert(key);
      localStorage[key] = value;
      return StorageManager;
    };

    StorageManager.get = function (key) {
      key = convert(key);
      return localStorage[key] ? localStorage[key] : null;
    };

    StorageManager.getStrings = function () {
      var list = [];
      for (var key in localStorage) {
        if (white_lists.indexOf(key) == -1) {
          list.push({
            name: key,
            src: localStorage[key]
          });
        }
      }
      return list;
    };

    StorageManager.on = function (key) {
      $('#' + key).on('keyup', function () {
        localStorage[convert(key)] = $(this).val();
      });
    };

    return StorageManager;
  })();

  //global variables
  var lock = (function () {
    var is_lock = false;

    return {
      isLock: function () {
        return is_lock;
      },
      lock: function () {
        is_lock = true;
      },
      unlock: function () {
        is_lock = false;
      }
    };
  })();

  var RESOURCES = {
    image_names: [],
    temp_image: 1,
    string_names: [],
  };

  //class
  var ProcessingManager = (function () {

    var RE_TO_MATCH = /\[(#+)\]\.png$/i
    , RE_FOR_REPLACE = /(\[#+\])(\.png)$/i;

    function ProcessingManager() {
      this.initialize.apply(this, arguments);
    }

    ProcessingManager.prototype = {
      constructor: ProcessingManager,
      initialize: function (canvas, $code, $folder, $subfolder, $file, $comp, $layer, $count, $mouse) {
        this.p = null;
        this.mouses = [];
        this.parameters = {};
        this.canvas = canvas;
        this.$code = $code;
        this.$folder = $folder;
        this.$subfolder = $subfolder;
        this.$file = $file;
        this.$comp = $comp;
        this.$layer = $layer;
        this.$count = $count;
        this.$mouse = $mouse;
        this.defaultFuncs = {
          multAlpha: (function (manager) {
            return function (value) {
              var p = this;
              value = parseFloat(value);
              if (!isNaN(value)) {
                value = clamp(value, 0, 1);
                var canvas = manager.canvas
                , ctx = canvas.getContext('2d');
                if (ctx) {
                  var width = p.width
                  , height = p.height
                  , data = ctx.getImageData(0, 0, width, height)
                  , pixels = data.data;
                  for(var y = 0; y < height; y++) {
                    for (var x = 0; x < width; x++) {
                      pixels[4 * (x + y * width) + 3] = Math.floor(pixels[4 * (x + y * width) + 3] * value);
                    }
                  }
                  ctx.putImageData(data, 0, 0);
                }
              }
            };
          })(this),
          glitch: (function (manager, base64_chars) {
            return function (type) {
              type = type || 'jpg';
              var p = this
              , canvas = manager.canvas
              , ctx = canvas.getContext('2d');
              if (ctx && (type == 'jpg' || type == 'png')) {
                p.noLoop();
                var data = type == 'jpg' ? canvas.toDataURL('image/jpeg').replace(/data:image\/jpeg;base64,/i, '') : canvas.toDataURL('image/png').replace(/data:image\/png;base64,/i, '')
                , c = base64_chars[~~(Math.random() * base64_chars.length)]
                , dst = type == 'jpg' ? "data:image/jpeg;base64," : "data:image/png;base64,";
                for (var ch in data) {
                  if(data[ch] == c && Math.random()<0.003){
                    dst += base64_chars[~~(Math.random() * base64_chars.length)];
                  }else{
                    dst += data[ch];
                  }
                }
                var image = new Image();
                image.onload = function(){
                  ctx.drawImage(image, 0, 0);
                  if (p && p.__is_alive) {
                    p.loop();
                  }
                };
                image.onerror = function () {
                  if (p && p.__is_alive) {
                    p.loop();
                  }
                };
                image.src = dst;
              }
            };
          })(this, 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'),
          zfill: zfill,
          toAE: (function (manager) {
            return function (key, value) {
              manager.parameters[key] = manager.parameters[key] || {};
              manager.parameters[key][this.frameCount] = value;
            };
          })(this),
        };
        this.funcs = {};
        this.fps = null;
      },
      draw: function (fn, onFail) {
        var self = this;
        if (lock.isLock() || this.p){
          if (onFail) {
            onFail();
          }
          return;
        }
        lock.lock();
        this.fps = null;
        this._replaceCanvas();
        self._isValid(self, fn, onFail).then(self._prepare).then(self._init).fail(self._handleError).always(self._doFinally);
      },
      _replaceCanvas: function () {
        $('#canvas').parent().empty().append('<canvas id="canvas"></canvas>');
        this.canvas = document.getElementById("canvas");
      },
      _isValid: function (self, fn, onFail) {
        var d = new $.Deferred();

        //folder check
        var folder_path
        , folder_name = folder_path = self._getFolderName();
        if (folder_name === '') {
          d.reject('Input a folder path.', onFail);
          return d.promise();
        }
        var folder = window.cep.fs.readdir(folder_name);
        if (folder.err != window.cep.fs.NO_ERROR) {
          d.reject('Folder not found.', onFail);
          return d.promise();
        }
        //subfolder check
        var subfolder_name = self._getSubFolderName();
        if (subfolder_name !== '') {
          if (subfolder_name.match(/\:|\;|\/|\||\,|\*|\?|\"|\'|\<|\>/) !== null) {
            d.reject('Invalid subfolder name.', onFail);
            return d.promise();
          }
          folder_path = folder_name + '/' + subfolder_name;
          var subfolder = window.cep.fs.readdir(folder_path);
          switch(subfolder.err) {
            case window.cep.fs.NO_ERROR:
              break;
            case window.cep.fs.ERR_NOT_FOUND:
              subfolder = window.cep.fs.makedir(folder_path);
              if (subfolder.err != window.cep.fs.NO_ERROR) {
                d.reject('File access denied.', onFail);
                return d.promise();
              }
              break;
            default:
              d.reject('File access denied.', onFail);
              return d.promise();
          }
        }

        //filename check
        var file_name = self._getFileName()
        , match = file_name.match(RE_TO_MATCH);
        if (match === null) {
          d.reject('Invalid file name.', onFail);
          return d.promise();
        }

        //confirm overwrite
        var width = match[1].length
        , image_sequence = folder_path + '/' + file_name.replace(RE_FOR_REPLACE, zfill(0, width) + '$2')
        , file = window.cep.fs.readFile(image_sequence);
        if (file.err == window.cep.fs.NO_ERROR) {
          var is_overwrite = confirm('Overwrite the image sequence?');
          if (!is_overwrite) {
            d.reject('', onFail);
            return d.promise();
          }
        }
        self._delete_images();

        //layer check
        var comp_name = self._getCompName()
        , layer_name = self._getLayerName();
        if (comp_name !== '' && layer_name !== '') {
          var data = {
            folder: folder_name,
            comp: comp_name,
            layer: layer_name,
          };
          new CSInterface().evalScript('$._ext.isExistLayer("' + encodeData(data) + '")', function (result) {
            //return value is not implemented yet
            var temp_file_name = folder_name + '/.temp'
            , temp_file = window.cep.fs.readFile(temp_file_name);
            if (temp_file.err == window.cep.fs.NO_ERROR) {
              window.cep.fs.deleteFile(temp_file_name);
              result = JSON.parse(temp_file.data);
              if (result.err === 0) {
                d.resolve(self, fn, onFail);
              } else {
                d.reject('Layer not found.', onFail);
              }
            } else {
              d.reject('File access denied.', onFail);
            }
          });
        } else {
          d.resolve(self, fn, onFail);
        }
        return d.promise();
      },
      _prepare: function (self, fn, onFail) {
        var d = new $.Deferred
        , folder_path = self._getFolderPath()
        , file_template = self._getFileName()
        , match = file_template.match(RE_TO_MATCH)
        , width = match[1].length
        , comp_name = self._getCompName()
        , layer_name = self._getLayerName();

        self.funcs = $.extend({}, self.defaultFuncs, {
          IMAGE_NAMES: RESOURCES.image_names.slice(0),
          loadFile: (function (folder) {
            return function (filename) {
              var path = folder + '/' + filename
              , file = window.cep.fs.readFile(path);
              if (file.err == window.cep.fs.NO_ERROR) {
                return file.data;
              } else {
                file = window.cep.fs.readFile(filename);
                if (file.err == window.cep.fs.NO_ERROR) {
                  return file.data;
                }
              }
              return null;
            };
          })(self._getFolderName()),
          loadObj: (function (folder) {
            function parse(text) {
              function getFace(match) {
                var indices = $.trim(match[1]).split(/\s+/)
                , vertices = []
                , textures = []
                , normals = []
                , face = {};
                for (var i = 0, l = indices.length; i < l; i++) {
                  var index = indices[i].split('/');
                  switch (index.length) {
                    case 1:
                      vertices.push(parseInt(index[0]) - 1);
                      break;
                    case 2:
                      vertices.push(parseInt(index[0]) - 1);
                      textures.push(parseInt(index[1]) - 1);
                      break;
                    case 3:
                      vertices.push(parseInt(index[0]) - 1);
                      if (index[1] !== '') {
                        textures.push(parseInt(index[1]) - 1);
                      }
                      normals.push(parseInt(index[2]) - 1);
                      break;
                  }
                }
                if (vertices.length > 0) {
                  face.vertices = vertices;
                }
                if (textures.length > 0) {
                  face.textures = textures;
                }
                if (normals.length > 0) {
                  face.normals = normals;
                }
                return face;
              }
              var lines = text.split('\n')
              , vertices = []
              , textures = []
              , normals = []
              , faces = []
              , match;
              for (var i = 0, l = lines.length; i < l; i++) {
                var line = $.trim(lines[i]);
                if (line === ''){
                  continue;
                } else if ((match = line.match(/v\s+([+-]?\d*\.?\d*)\s+([+-]?\d*\.?\d*)\s+([+-]?\d*\.?\d*)/)) !== null) {
                  vertices.push([parseFloat(match[1]), parseFloat(match[2]), parseFloat(match[3])]);
                } else if ((match = line.match(/vt\s+([+-]?\d*\.?\d*)\s+([+-]?\d*\.?\d*)/)) !== null) {
                  textures.push([parseFloat(match[1]), parseFloat(match[2])]);
                } else if ((match = line.match(/vn\s+([+-]?\d*\.?\d*)\s+([+-]?\d*\.?\d*)\s+([+-]?\d*\.?\d*)/)) !== null) {
                  normals.push([parseFloat(match[1]), parseFloat(match[2]), parseFloat(match[3])]);
                } else if ((match = line.match(/f((?:\s+\d+){3,})/)) !== null) {
                  faces.push(getFace(match));
                } else if ((match = line.match(/f((?:\s+\d+\/\d+){3,})/)) !== null) {
                  faces.push(getFace(match));
                } else if ((match = line.match(/f((?:\s+\d+\/\/\d+){3,})/)) !== null) {
                  faces.push(getFace(match));
                } else if ((match = line.match(/f((?:\s+\d+\/\d+\/\d+){3,})/)) !== null) {
                  faces.push(getFace(match));
                }
              }
              return {
                vertices: vertices,
                textures: textures,
                normals: normals,
                faces: faces,
              };
            }
            return function (filename) {
              var path = folder + '/' + filename
              , file = window.cep.fs.readFile(path);
              if (file.err == window.cep.fs.NO_ERROR) {
                return parse(file.data);
              } else {
                file = window.cep.fs.readFile(filename);
                if (file.err == window.cep.fs.NO_ERROR) {
                  return parse(file.data);
                }
              }
              return {
                vertices: [],
                faces: [],
              };
            };
          })(self._getFolderName()),
        });

        if (comp_name !== '' && layer_name !== '') {
          var data = {
            folder: folder_path,
            comp: comp_name,
            layer: layer_name,
          };
          new CSInterface().evalScript('$._ext.getLayerControls("' + encodeData(data) + '")', function (result) {
            //return value is not implemented yet
            var temp_file_name = folder_path + '/.temp';
            var temp_file = window.cep.fs.readFile(temp_file_name);
            if (temp_file.err == window.cep.fs.NO_ERROR) {
              window.cep.fs.deleteFile(temp_file_name);
              result = JSON.parse(temp_file.data);
              if (result.err === 0) {
                var data = result.data;
                self.fps = data.fps;
                var frames = data.frames;
                //__after
                self.funcs['__after'] = (function(manager, folder, file, width, canvas, frames) {
                  var max_frame = ~~new Array(width + 1).join('9')
                  , date = new Date()
                  , txt = ''
                  ;
                  return function () {
                    manager.mouses.push([this.mouseX, this.mouseY]);
                    var frame_count = this.frameCount
                    , frame = zfill(frame_count, width)
                    , file_name = folder + '/' + file.replace(RE_FOR_REPLACE, frame + '$2')
                    , data = canvas.toDataURL().replace(/data:image\/png;base64,/i, '')
                    , new_date = new Date()
                    , time = new_date - date
                    ;
                    window.cep.fs.writeFile(file_name, data, window.cep.encoding.Base64);
                    date = new_date;
                    if (frame_count === 0) {
                      txt = frame;
                    } else {
                      txt = frame + ' (' + (time / 1000).toFixed(3) + 's)';
                    }
                    manager.$count.text(txt);
                    if (frame_count >= frames - 1 || frame_count >= max_frame) {
                      this.exit();
                    }
                  };
                })(self, folder_path, file_template, width, self.canvas, frames);
                //controls
                var values = data.values;
                for (var control in values) {
                  self.funcs[control] = (function(array) {
                    var len = array.length;
                    return function (frame) {
                      frame = parseInt(frame, 10);
                      if (isNaN(frame)) {
                        frame = this.frameCount;
                      } else {
                        if (frame < 0) {
                          frame = this.frameCount - frame;
                        }
                      }
                      if (frame < 0) {
                        return array[0];
                      } else if (frame < len) {
                        return array[frame];
                      }
                      return array[len - 1];
                    };
                  })(values[control]);
                }
                d.resolve(self, fn, onFail);
              } else {
                d.reject('Layer not found.', onFail);
              }
            } else {
              d.reject('File access denied.', onFail);
            }
          });
        } else {
          //__after
          self.funcs['__after'] = (function (manager, folder, file, width, canvas) {
            var max_frame = ~~new Array(width + 1).join('9')
            , date = new Date()
            , txt = ''
            ;
            return function () {
              manager.mouses.push([this.mouseX, this.mouseY]);
              var frame_count = this.frameCount
              , frame = zfill(frame_count, width)
              , file_name = folder + '/' + file.replace(RE_FOR_REPLACE, frame + '$2')
              , data = canvas.toDataURL().replace(/data:image\/png;base64,/i, '')
              , new_date = new Date()
              , time = new_date - date
              ;
              window.cep.fs.writeFile(file_name, data, window.cep.encoding.Base64);
              date = new_date;
              if (frame_count === 0) {
                txt = frame;
              } else {
                txt = frame + ' (' + (time / 1000).toFixed(3) + 's)';
              }
              manager.$count.text(txt);
              if (frame_count >= max_frame) {
                this.exit();
              }
            };
          })(self, folder_path, file_template, width, self.canvas);
          d.resolve(self, fn, onFail);
        }
        return d.promise();
      },
      _init: function (self, fn, onFail) {
        var d = new $.Deferred();
        try {
          self.p = new Processing(self.canvas, self.$code.val(), self.funcs, fn);
          d.resolve(self);
        } catch (e) {
          d.reject('Invalid code: ' + e.message, onFail);
        }
        return d.promise();
      },
      _handleError: function (e, onFail) {
        if (onFail) {
          onFail();
        }
        if (e !== '') {
          alert(e);
        }
      },
      _doFinally: function () {
        lock.unlock();
      },
      redraw: function (fn, onFail) {
        var self = this;
        if (lock.isLock()){
          return;
        }
        lock.lock();

        this.quit();
        setTimeout(function () {
          self._init(self, fn, onFail).fail(self._handleError).always(self._doFinally);
        }, 500);
      },
      quit: function () {
        this._destructor();
        this._delete_images();
      },
      ok: function () {
        if (this.p) {
          //add item
          var folder_path = this._getFolderPath()
          , file_template = this._getFileName()
          , match = file_template.match(RE_TO_MATCH)
          , width = match[1].length
          , file_name = folder_path + '/' + file_template.replace(RE_FOR_REPLACE, zfill(0, width) + '$2')
          , framerate = this.fps || this.p.getFrameRate();

          var data = {
            filename: file_name,
            comp: this._getCompName(),
            layer: this._getLayerName(),
            framerate: framerate,
            isMouse: this._isMouseChecked(),
            mouses: this.mouses,
            parameters: this.parameters,
          };

          new CSInterface().evalScript('$._ext.addSequence("' + encodeData(data) + '")');
          //destruct
          this._destructor();
        }
      },
      _destructor: function () {
        if (this.p) {
          this.p.exit();
          this.p = null;
        }
        this.mouses = [];
        this.parameters = {};
      },
      _delete_images: function () {
        if (this.p) {
          this.p.noLoop();
        }
        var data = {
          folder: this._getFolderPath(),
          filename: this._getFileName()
        };
        new CSInterface().evalScript('$._ext.deleteFiles("' + encodeData(data) + '")');
      },
      _getFolderPath: function () {
        var folder_name = this._getFolderName();
        var subfolder_name = this._getSubFolderName();
        return subfolder_name != '' ? folder_name + '/' + subfolder_name : folder_name;
      },
      _getFolderName: function () {
        return this.$folder.val();
      },
      _getSubFolderName: function () {
        return this.$subfolder.val();
      },
      _getFileName: function () {
        return this.$file.val();
      },
      _getCompName: function () {
        return this.$comp.val();
      },
      _getLayerName: function () {
        return this.$layer.val();
      },
      _isMouseChecked: function () {
        return this.$mouse.prop('checked');
      }
    };

    return ProcessingManager;
  })();

  /**
  * functions for UI
  */
  function isCanvasActive() {
    return $('.canvas_wrapper').hasClass('active');
  }

  function getFolderPath() {
    return $('#folder').val();
  }

  function tabToSpaces(e) {
    var space_num = 2;
    if (e.keyCode === 9) {
      e.preventDefault();
      var elem = e.target
      , val = elem.value
      , pos = elem.selectionStart;
      if(e.shiftKey){
        for(var i = 0, p = pos - 1; i < space_num && p >= 0; i++, p--){
          if(val[p] !== ' ') {
            break;
          }
        }
        if(i >= 0){
          elem.value = val.substr(0, pos - i) + val.substr(pos, val.length);
          elem.setSelectionRange(pos - i, pos - i);
        }
      }else{
        elem.value = val.substr(0, pos) + '  ' + val.substr(pos, val.length);
        elem.setSelectionRange(pos + space_num, pos + space_num);
      }
    }
  }

  //resources
  function updateResourcesInfo() {
    $('#images_info').html(RESOURCES.image_names.map(function (str) { return '<span class="loaded_image">' + str + '</span>' }).join('<br>'));
    $('#strings_info').html(RESOURCES.string_names.map(function (str) { return '<span class="loaded_string">' + str + '</span>' }).join('<br>'));
  }

  function initializeImages() {
    $('.image_tooltip').remove();
    RESOURCES.image_names = [];
    RESOURCES.temp_image = 1;
    updateResourcesInfo();
  }

  function initializeStrings() {
    StorageManager.clear();
    RESOURCES.string_names = [];
    updateResourcesInfo();
  }

  function removeImage(name) {
    var tooltip = $('.image_tooltip[name="' + name + '"]');
    tooltip.remove();
    RESOURCES.image_names = RESOURCES.image_names.filter(function (v) { return v != name;});
    updateResourcesInfo();
  }

  function removeString(name) {
    delete localStorage[name];
    RESOURCES.string_names = RESOURCES.string_names.filter(function (v) { return v != name;});
    updateResourcesInfo();
  }

  function appendResources(resources, total, fn) {
    function next (resources, total, fn) {
      if (resources.length > 0) {
        appendResources(resources, total, fn);
      } else {
        $('#resource_loading').css('width', '0');
        updateResourcesInfo();
        if (fn) {
          fn();
        }
      }
    }

    //progress bar
    $('#resource_loading').css('width', (100 * (total - resources.length) / total) + '%');

    var data = resources.shift();
    if (!data) {
      next(resources, total, fn);
    }

    if (data.type == 'image') {
      var name = data.name
      , src = data.src;

      var img = new Image();
      img.onload = function () {
        var div = document.createElement('div');
        div.className = 'image_tooltip';
        div.setAttribute('name', name);
        div.appendChild(img);
        var span = document.createElement('span');
        span.textContent = img.width + 'x' + img.height;
        div.appendChild(span);
        var body = document.getElementsByTagName('body')[0];
        body.appendChild(div);
        RESOURCES.image_names.push(name);
        next(resources, total, fn);
      }
      img.onerror = function () {
        next(resources, total, fn);
      };
      img.name = name;
      img.src = src;
    } else if (data.type == 'string') {
      var name = data.name
      , src = data.src;
      localStorage[name] = src;
      RESOURCES.string_names.push(name);
      next(resources, total, fn);
    } else {
      next(resources, total, fn);
    }
  }

  function loadResources(options) {
    function loadImage(file_name) {
      var name = /.*\/(.+?)$/.exec(file_name)[1]
      if (RESOURCES.image_names.indexOf(name) >= 0) {
        return null;
      }
      return {type: 'image', name: name, src: file_name};
    }

    function loadString(file_name) {
      var name = /.*\/(.+?)$/.exec(file_name)[1]
      if (RESOURCES.string_names.indexOf(name) >= 0) {
        return null;
      }
      var file = window.cep.fs.readFile(file_name);
      if (file.err == window.cep.fs.NO_ERROR) {
        var data = file.data;
        return {type: 'string', name: name, src: data};
      }
      return null;
    }

    if (isCanvasActive() || lock.isLock()) {
      return;
    }

    lock.lock();

    if (options) {
      options = $.extend({}, {is_folder: false, folder_path: ''}, options);
    } else {
      options = $.extend({}, {is_folder: false, folder_path: ''});
    }

    var is_folder = options.is_folder
    , folder_path = options.folder_path
    , files = null;
    if (is_folder) {
      if (folder_path == '') {
        var result = window.cep.fs.showOpenDialog(false, true, 'Select Folder.');
        if (result.err == window.cep.fs.NO_ERROR) {
          folder_path = result.data + '/';
          var folder = window.cep.fs.readdir(folder_path);
          if (folder.err == window.cep.fs.NO_ERROR) {
            files = folder.data.map(function (file_name) { return folder_path + file_name;});
          }
        }
      } else {
        var folder = window.cep.fs.readdir(folder_path);
        if (folder.err == window.cep.fs.NO_ERROR) {
          files = folder.data.map(function (file_name) { return folder_path + file_name;});
        }
      }
    } else {
      var result = window.cep.fs.showOpenDialog(true, false, 'Select image files.', '', ['gif', 'jpg', 'jpeg', 'png', 'txt', 'svg', 'dat']);
      if (result.err == window.cep.fs.NO_ERROR) {
        files = result.data;
      }
    }
    if (files) {
      var resources = [];
      for (var i = 0, l = files.length; i < l; i++) {
        var file = files[i]
        , obj = null;
        if (/\.(gif|jpg|jpeg|png)$/i.exec(file)) {
          obj = loadImage(file);
        } else if (/\.(txt|svg|dat)$/i.exec(file)) {
          obj = loadString(file);
        }
        if (obj) {
          resources.push(obj);
        }
      }
      appendResources(resources, resources.length, function () {lock.unlock()});
    } else {
      lock.unlock();
    }
  }

  function addImage() {
    var folder_path = getFolderPath();
    if (folder_path == '') {
      alert('Input a folder path.');
      return;
    }

    if (isCanvasActive() || lock.isLock()) {
      return;
    }

    lock.lock();

    var data = {
      folder: folder_path,
    };
    new CSInterface().evalScript('$._ext.addImage("' + encodeData(data) + '")', function (result) {
      //return value is not implemented yet
      try {
        var temp_file_name = folder_path + '/.temp'
        , temp_file = window.cep.fs.readFile(temp_file_name);
        if (temp_file.err == window.cep.fs.NO_ERROR) {
          window.cep.fs.deleteFile(temp_file_name);
          result = JSON.parse(temp_file.data);
          if (result.err == 0) {
            var img_file_path = result.data
            , file = window.cep.fs.readFile(img_file_path, window.cep.encoding.Base64);
            if (file.err == window.cep.fs.NO_ERROR) {
              window.cep.fs.deleteFile(img_file_path);
              var data = 'data:image/png;base64,' + file.data
              , name = '_temp' + RESOURCES.temp_image++ + '.png';
              while (RESOURCES.image_names.indexOf(name) >= 0) {
                name = '_temp' + RESOURCES.temp_image++ + '.png';
              }
              appendResources([{type: 'image', name: name, src: data}], 1, function () {lock.unlock();});
            } else {
              throw new Error('File access denied.');
            }
          } else {
            throw new Error('Activate a comp.');
          }
        } else {
          throw new Error('File access denied.');
        }
      } catch (e) {
        lock.unlock();
        alert(e);
      }
    });
  }

  function addImages() {
    var folder_path = getFolderPath();
    if (folder_path == '') {
      alert('Input a folder path.');
      return;
    }

    var subfolder = prompt("Input a subfolder name.");
    if (subfolder == null || subfolder == '' || subfolder.match(/\:|\;|\/|\||\,|\*|\?|\"|\'|\<|\>/) != null) {
      return;
    }

    if (isCanvasActive() || lock.isLock()) {
      return;
    }
    lock.lock();
    $('.loading_wrapper').addClass('active');

    var data = {
      folder: folder_path,
      subfolder: subfolder,
    };
    new CSInterface().evalScript('$._ext.addImages("' + encodeData(data) + '")', function (result) {
      //return value is not implemented yet
      try {
        var temp_file_name = folder_path + '/.temp'
        , temp_file = window.cep.fs.readFile(temp_file_name);
        if (temp_file.err == window.cep.fs.NO_ERROR) {
          window.cep.fs.deleteFile(temp_file_name);
          result = JSON.parse(temp_file.data);
          if (result.err == 0) {
            lock.unlock();
            $('.loading_wrapper').removeClass('active');
            loadResources({is_folder: true, folder_path: folder_path + '/' + subfolder + '/'});
          } else {
            throw new Error(result.data);
          }
        } else {
          throw new Error('File access denied.');
        }
      } catch (e) {
        lock.unlock();
        $('.loading_wrapper').removeClass('active');
        alert(e);
      }
    });
  }

  function switchLoadResources(e) {
    if (e.shiftKey) {
      loadResources({is_folder:true});
    } else {
      loadResources();
    }
  }

  function importPde() {
    if (isCanvasActive()) {
      return;
    }

    var result = window.cep.fs.showOpenDialog(false, false, 'Select a pde file.', '', ['pde', 'txt']);
    if (result.err == window.cep.fs.NO_ERROR) {
      var file_name = result.data[0];
      if (/\.(pde|txt)$/i.exec(file_name)) {
        var file = window.cep.fs.readFile(result.data[0]);
        if (file.err == window.cep.fs.NO_ERROR) {
          $('#code').val(file.data);
        }
      }
    }
  }

  function importProject() {
    if (isCanvasActive() || lock.isLock()) {
      return;
    }

    var result = window.cep.fs.showOpenDialog(false, false, 'Select a proejct file.', '', ['json']);
    if (result.err == window.cep.fs.NO_ERROR) {
      var file_name = result.data[0];
      if (/\.(json)$/i.exec(file_name)) {
        var file = window.cep.fs.readFile(result.data[0]);
        if (file.err == window.cep.fs.NO_ERROR) {
          var prject, is_valid = true;
          try {
            project = JSON.parse(file.data)
          } catch (e) {
            is_valid = false;
          }
          if (is_valid) {
            var resources = project.resources || [];
            lock.lock();
            //clear
            initializeImages();
            initializeStrings();
            //set current temp
            var temp_image = -1;
            for (var i = 0, l = resources.length; i < l; i++) {
              var resource = resources[i];
              if (resource.type == 'image' && typeof resource.name == 'string') {
                var match = resource.name.match(/^_temp(\d+)\.png$/);
                if (match != null) {
                  temp_image = Math.max(temp_image, ~~match[1]);
                }
              }
            }

            if (temp_image == -1) {
              RESOURCES.temp_image = 1;
            } else {
              RESOURCES.temp_image = temp_image + 1;
            }
            //load
            if (project.folder) {
              $('#folder').val(project.folder);
            }
            if (project.subfolder) {
              $('#subfolder').val(project.subfolder);
            }
            if (project.filename) {
              $('#file_name').val(project.filename);
            }
            if (project.comp) {
              $('#comp').val(project.comp);
            }
            if (project.layer) {
              $('#layer').val(project.layer);
            }
            if (project.code) {
              $('#code').val(project.code);
            }
            appendResources(resources, resources.length, function () {lock.unlock()});
          }
        }
      }
    }
  }

  function exportPde() {
    var code = $('#code').val();
    if (!isCanvasActive() && code != '') {
      new CSInterface().evalScript('$._ext.exportPde("' + encode(code)+ '")');
    }
  }

  function exportProject() {
    if (isCanvasActive()) {
      return;
    }

    var project = {
      folder: $('#folder').val(),
      subfolder: $('#subfolder').val(),
      filename: $('#file_name').val(),
      comp: $('#comp').val(),
      layer: $('#layer').val(),
      code: $('#code').val(),
      resources: [],
      VERSION: '1.1',
    };

    //images
    $('img').each(function () {
      var name = $(this).attr('name')
      , src = $(this).attr('src');
      project.resources.push({
        type: 'image',
        name: name,
        src: src,
      });
    });

    //strings
    var strings = StorageManager.getStrings();
    for (var i = 0, l = strings.length; i < l; i++) {
      var string = strings[i];
      project.resources.push({
        type: 'string',
        name: string.name,
        src: string.src,
      });
    }

    //export
    new CSInterface().evalScript('$._ext.exportProject("' + encode(JSON.stringify(project))+ '")');
  }

  function selectFolder() {
    var result = window.cep.fs.showOpenDialog(false, true, 'Select a folder.');
    if (result.err == window.cep.fs.NO_ERROR && result.data != '') {
      $('#folder').val(result.data);
      $('#folder').trigger('keyup');
    }
  }

  function pickLayer() {
    var folder_path = getFolderPath();
    if (folder_path == '') {
      alert('Input a folder path.');
      return;
    }

    var folder = window.cep.fs.readdir(folder_path);

    if (folder.err != window.cep.fs.NO_ERROR) {
      alert('Folder not found.');
      return;
    }

    if (isCanvasActive() || lock.isLock()) {
      return;
    }
    lock.lock();

    var data = {
      folder: folder_path,
    };
    new CSInterface().evalScript('$._ext.getActiveLayer("' + encodeData(data) + '")', function (result) {
      //return value is not implemented yet
      try {
        var temp_file_name = folder_path + '/.temp'
        , temp_file = window.cep.fs.readFile(temp_file_name);
        if (temp_file.err == window.cep.fs.NO_ERROR) {
          window.cep.fs.deleteFile(temp_file_name);
          result = JSON.parse(temp_file.data);
          if (result.err == 0) {
            var data = result.data;
            $('#comp').val(data.comp_name);
            $('#layer').val(data.layer_name);
          } else {
            throw new Error(result.data);
          }
        } else {
          throw new Error('File access denied.');
        }
      } catch (e) {
        if (e.message != '') {
          alert(e);
        }
      } finally {
        lock.unlock();
      }
    });
  }

  function clearLayer() {
    $('#comp').val('');
    $('#layer').val('');
  }

  function draw() {
    //loading
    $('.loading_wrapper').addClass('active');
    p.draw(
      function () {
        $('.loading_wrapper').removeClass('active');
        $('.canvas_wrapper').addClass('active');
      },
      function () {
        $('.loading_wrapper').removeClass('active');
      }
    );
  }

  function forceDraw() {
    $('.loading_wrapper').addClass('active');
    p.redraw(
      function () {
        $('.loading_wrapper').removeClass('active');
        $('.canvas_wrapper').addClass('active');
      },
      function () {
        $('.loading_wrapper').removeClass('active');
      }
    );
  }

  function redraw() {
    p.redraw();
  }

  function quit() {
    p.quit();
    $('.canvas_wrapper').removeClass('active');
  }

  function ok() {
    p.ok();
    $('.canvas_wrapper').removeClass('active');
  }

  function checkMouse() {
    if ($('#mouse').prop('checked')) {
      $('#mouse').prop('checked', false);
    } else {
      $('#mouse').prop('checked', true);
    }
  }

  function initializeInput(objs) {
    var obj = objs.shift()
    , key = obj.key
    , default_value = obj.default_value || '';

    var value = StorageManager.get(key) ? StorageManager.get(key) : default_value;

    $('#' + key).val(value);
    StorageManager.on(key);

    if (objs.length > 0) {
      initializeInput(objs);
    }
  }

  function initializeCode() {
    var default_code = [
      '//code from http://processingjs.org/learning/',
      '',
      'float radius = 50.0;',
      'int X, Y;',
      'int nX, nY;',
      'int delay = 16;',
      '',
      'void setup(){',
      '  size(200, 200);',
      '  strokeWeight(10);',
      '  frameRate(30);',
      '  X = width / 2;',
      '  Y = height / 2;',
      '  nX = X;',
      '  nY = Y;',
      '}',
      '',
      'void draw(){',
      '  radius = radius + sin(frameCount / 4);',
      '  X += (nX - X) / delay;',
      '  Y += (nY - Y) / delay;',
      '  background(100);',
      '  fill(0, 121, 184);',
      '  stroke(255);',
      '  ellipse(X, Y, radius, radius);',
      '}',
      '',
      'void mouseMoved(){',
      '  nX = mouseX;',
      '  nY = mouseY;',
      '}'
    ].join('\n');

    var $code_index
    , $save_code
    , $delete_code
    , $code;

    function init() {
      $code_index = $('#code_index');
      $save_code = $('#save_code');
      $delete_code = $('#delete_code');
      $code = $('#code');

      setEvents();
      refreshUI();
      setCode();
    }

    function setEvents() {
      $code_index.on('change', setCode);
      $save_code.on('click', function (e) {
        if (e.shiftKey) {
          overwriteCode();
        } else {
          saveCode();
        }
      });
      $delete_code.on('click', deleteCode);
    }

    function setCode() {
      var index = $code_index.val()
      , values = getStorage();

      if (index >= 0 && index < values.length) {
        $code.val(values[index].code);
      } else {
        alert("Error.");
        refreshUI();
      }
    }

    function saveCode(){
      var name = prompt("Input a code name.");
      if (name != null) {
        var values = getStorage();
        values.push({
          name: name,
          code: $code.val()
        });
        setStorage(values);
        refreshUI();
        setIndex(values.length - 1);
      }
    }

    function overwriteCode(){
      var index = $code_index.val()
      , values = getStorage()
      , is_overwrite = confirm('Overwrite "' + values[index].name + '"?');
      if (is_overwrite) {
        var value = {
          name: values[index].name,
          code: $code.val()
        };
        values.splice(index, 1, value);
        setStorage(values);
        refreshUI();
        setIndex(index);
      }
    }

    function deleteCode() {
      var index = $code_index.val()
      , values = getStorage()
      , is_delete = confirm('Delete "' + values[index].name + '"?');
      if (is_delete) {
        values.splice(index, 1);
        setStorage(values);
        refreshUI();
        setIndex(Math.min(index, values.length - 1));
        setCode();
      }
    }

    function getStorage() {
      try {
        var value = StorageManager.is('code') ? JSON.parse(StorageManager.get('code')) : null;
      } catch (e) {
        value = null;
      }
      if (!value || !(value instanceof Array) || value.length == 0) {
        var values = [];
        values.push({
          name: 'default',
          code: default_code
        });
        setStorage(values);
      }

      return JSON.parse(StorageManager.get('code'));
    }

    function setStorage(values) {
      StorageManager.set('code', JSON.stringify(values))
    }

    function setIndex(index) {
      $code_index.val(index);
    }

    function refreshUI() {
      $code_index.empty();
      var values = getStorage()
      , fragment = document.createDocumentFragment();
      for (var i = 0, l = values.length; i < l; i++) {
        var obj = values[i]
        , option = document.createElement('option');
        option.setAttribute('value', i);
        option.textContent = obj.name;
        fragment.appendChild(option);
      }
      $code_index.append(fragment);
    }

    init();
  }

  //processing mangaer
  var p;

  function init () {
    //ui
    $('textarea').on('keydown', tabToSpaces);
    $('#add_image').on('click', function (e) {
      if(e.shiftKey) {
        var is_add = confirm('Add images?');
        if (is_add) {
          addImages();
        }
      } else {
        addImage();
      }
    });
    $('#load_resources').on('click', switchLoadResources);
    $('#import').on('click', function (e) {
      if (e.shiftKey) {
        importProject();
      } else {
        importPde();
      }
    });
    $('#export').on('click', function (e) {
      if (e.shiftKey) {
        exportProject();
      } else {
        exportPde();
      }
    });
    $('#folder').on('click', selectFolder);
    $('#clear').on('click', clearLayer);
    $('#pick').on('click', pickLayer);
    $('#draw').on('click', function (e) {
      if(e.shiftKey) {
        var is_force = confirm('Ignore initializations?');
        if (is_force) {
          forceDraw();
        }
      } else {
        draw();
      }
    });
    $('#redraw').on('click', redraw);
    $('#quit').on('click', quit);
    $('#ok').on('click', ok);
    $('#images_info').on('click', function (e) {
      if(e.shiftKey) {
        var is_initialize = confirm('Initialize images?');
        if (is_initialize) {
          initializeImages();
        }
      }
    });
    $('#strings_info').on('click', function (e) {
      if(e.shiftKey) {
        var is_initialize = confirm('Initialize strings?');
        if (is_initialize) {
          initializeStrings();
        }
      }
    });

    //tooltip
    $(document).on({
      contextmenu: function (e) {
        e.preventDefault();
        var name = $(this).text();
        var is_remove = confirm('Remove ' + name + '?');
        if (is_remove) {
          removeImage(name);
        }
      },
      mouseenter: function (e) {
        var name = $(this).text()
        , tooltip = $('.image_tooltip[name="' + name + '"]')
        , w = tooltip.innerWidth()
        , h = tooltip.innerHeight()
        , x = e.pageX
        , y = e.pageY;
        tooltip.addClass('active').css({left:x + 20, top: y - h - 10});
      },
      mousemove: function (e) {
        var name = $(this).text()
        , tooltip = $('.image_tooltip[name="' + name + '"]')
        , w = tooltip.innerWidth()
        , h = tooltip.innerHeight()
        , x = e.pageX
        , y = e.pageY;
        tooltip.css({left:x + 20, top: y - h - 10});
      },
      mouseleave: function (e) {
        $('.image_tooltip.active').removeClass('active');
      }
    }, '.loaded_image');

    $(document).on({
      contextmenu: function (e) {
        e.preventDefault();
        var name = $(this).text();
        var is_remove = confirm('Remove ' + name + '?');
        if (is_remove) {
          removeString(name);
        }
      },
      click: function (e) {
        if ($('.string_tooltip').length > 0) {
          return;
        }

        var name = $(this).text()
        , tooltip = $('<div/>', {class: 'string_tooltip'})
        , html
        , type = /\.(svg|txt|dat)$/i.exec(name)[1].toLowerCase();
        switch (type) {
          case 'svg':
            tooltip.css({'width': 300, 'height': 200, 'background-image': 'url(data:image/svg+xml;base64,' + window.cep.encoding.convertion.utf8_to_b64(localStorage[name]) + ')'});
            break;
          case 'txt':
            html = localStorage[name].split('\n').map(function (str, i) {return i + ': ' + htmlspecialchars(str);}).join('<br>');
            break;
          case 'dat':
            html = Array.prototype.slice.call(localStorage[name], 0, 1000).map(function (str, i) {var s = str.charCodeAt(0); var r = i % 20; return r == 19 ? ',' + s + '<br>' : r > 0 ? ',' + s : s;}).join('');
            break;
          default:
            return;
        }
        tooltip.html(html);
        $('body').append(tooltip);
        var w = tooltip.innerWidth()
        , h = tooltip.innerHeight()
        , x = e.pageX
        , y = e.pageY;
        switch (type) {
          case 'svg':
            tooltip.css({left: x - 0.1 * w, top: y - 0.75 * h});
            break;
          case 'txt':
            tooltip.css({left: x - 0.1 * w, top: y - 0.25 * h});
            break;
          case 'dat':
            tooltip.css({left: x - 0.1 * w, top: y - 0.25 * h});
            break;
        }
      }
    }, '.loaded_string');

    $(document).on({
      mouseleave: function (e) {
        $('.string_tooltip').remove();
      }
    }, '.string_tooltip');

    //shortcut key
    $(window).on('keydown', function (e) {
      if (!isCanvasActive()) {
        if (e.ctrlKey) {
          switch (e.keyCode) {
            case 68://d
              return draw();
          }
        }
      } else {
        //esc
        if (e.keyCode == 27) {
          return quit();
        }
        if (e.ctrlKey) {
          switch (e.keyCode) {
            case 69://e
              return ok();
            case 77://m
              return checkMouse();
            case 81://q
              return quit();
            case 82://r
              return redraw();
          }
        }
      }
    });

    StorageManager.clear();
    initializeCode();
    initializeInput([
      {key: 'folder'},
      {key: 'subfolder'},
      {key: 'file_name', default_value: 'processing_[#####].png'}
    ]);

    loadJSX();
    p = new ProcessingManager(document.getElementById("canvas"), $('#code'), $('#folder'), $('#subfolder'), $('#file_name'), $('#comp'), $('#layer'), $('#count'), $('#mouse'));
  }

  init();
});
